/******************************************************************************
HLSL Eye Shader
by Rob Galanakis
robg@robg3d.com
www.robg3d.com

The eye shader is designed around two philosophies seemingly at odds: visual fidelity, and 
variation with low overhead. The same eye shader and set of textures provides 4 eye "types" with an 
infinite number of colors. Adding a single 256x256 RGBA texture provides a whole new set of 4 eye types. 
So, instead of having 10 eye textures for 100 characters and having repeating eyes, the same 10 textures 
can produce 40 100% unique eyes, and the rest of them with 100% unique eye colours.

The variation is achieved with an RGB texture that does not control diffuse colour, but are masks for 
externally-set diffuse colours. The inner iris, outer iris, and veins all have a unique colour that can 
be set differently for every single eye with absolutely no hit whatsoever to performance. On top of this 
"colour comp" is applied one of four "overlay" channels from an RGBA texture to give finer detail.

Visual fidelity is handled in two parts. The first, is the cornea (the whites of the eye). The ambient 
colour controls the overall reflectivity of the unlit areas of the eye. There is a variable for biasing 
the diffuse lighting. A normal map creates the "bulge" of the cornea over the front of the eye and iris.
 Most importantly, the specular highlights are multiplied by a reflection map, which breaks up the 
 specular highlight. Because eyes are 100% reflective, they clearly reflect the shape of their light 
 source. Thus, the traditional round shape of the specular highlight looks "fake." The broad specular 
 broken up by the reflection map (which has brightness and contrast controls) approximates an 
 image-based lighting approach very cheaply.

The second part is the iris. The iris' specular highlight is broad, and takes the colour of the iris. 
The inverted normal map is used to approximate the "inverted hemisphere" of the iris, that recedes from 
the eyeball's spherical surface.

rgbTexture: R channel holds the vein color mask, G holds the inner iris colour mask, B holds the outer 
iris colour mask.

overlayTexture: Each channel (RGBA) holds an overlay texture that is multiplied over the eye's base 
colour (after the application of the eye colour variables to the rgbTexture masks).

normalTexture: Should have a "bulge" out in the iris area.

reflTexture: The cube map that is the ambient reflection and breaks up the specular highlight. For best 
results, should have areas of strong contrast, and irregular shapes (such as, a forest).

******************************************************************************/

 //Define textures
texture rgbTexture : DIFFUSE
<
	string ResourceName = "eye_rgb.tga";
	string ResourceType = "2D";
>;

texture overlayTexture : DIFFUSE
<
	string ResourceName = "eye_overlay.tga";
	string ResourceType = "2D";
>;

texture normalTexture : NORMAL
<
	string ResourceName = "eye_normal.tga";
	string ResourceType = "2D";
>;

texture reflTexture : REFLECTION
<
	string ResourceName = "c_foliage_16M.dds";
	string ResourceType = "cube";
>;

//Texture samplers

sampler2D texSamplerRGB = sampler_state
{
    texture = <rgbTexture>;
    MIPFILTER = LINEAR;
    MINFILTER = LINEAR;
    MAGFILTER = LINEAR;
};

sampler2D texSamplerOverlay = sampler_state
{
	Texture = <overlayTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
};

sampler2D texSamplerNormal = sampler_state
{
	Texture = <normalTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
};

samplerCUBE texSamplerRefl = sampler_state
{
	Texture = <reflTexture>;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
};

//Automatic variables

float4x4 wvp : WorldViewProjection < string UIWidget = "None"; >;  
float4x4 worldI : WorldInverse < string UIWidget = "None"; >;  
float4x4 worldIT : WorldInverseTranspose < string UIWidget = "None"; >;  
float4x4 viewInv : ViewInverse < string UIWidget = "None"; >;  
float4x4 world : World < string UIWidget = "None"; >;

//Tweakables
//This controls how much or how little diffuse lighting there is
float diffuseBias	: SPECULAR
<
	string UIWidget = "slider";
	float UIMin = -1.0;
	float UIMax = 1.0;
	float UIStep = 0.05;
	string UIName = "diffuseBias";
> = 0.0;

//Controls the brightness of the reflection cube map
float reflBrightness	: SPECULAR
<
	string UIWidget = "slider";
	float UIMin = -2.0;
	float UIMax = 2.0;
	float UIStep = 0.05;
	string UIName = "reflBrightness";
> = 0.0;
//Controls the contrast of the reflection cube map
float reflContrast	: SPECULAR
<
	string UIWidget = "slider";
	float UIMin = 0.0;
	float UIMax = 15.0;
	float UIStep = 0.1;
	string UIName = "reflContrast";
> = 1.0;

//The size of the specular of the whites of the eye
float glossCornea	: SPECULARPOWER
<
	string UIWidget = "slider";
	float UIMin = 1.0;
	float UIMax = 128.0;
	float UIStep = 1.0;
	string UIName = "glossCornea";
> = 32.0;
//size of the specular of the iris
float glossIris	: SPECULARPOWER
<
	string UIWidget = "slider";
	float UIMin = 1.0;
	float UIMax = 128.0;
	float UIStep = 1.0;
	string UIName = "glossIris";
> = 5.0;

//Controls which "overlay" texture is used
float alphaPicker	: DIFFUSE
<
	string UIWidget = "slider";
	float UIMin = 0.0;
	float UIMax = 3.0;
	float UIStep = 1.0;
	string UIName = "alphaPicker";
> = 0.0;

//Color of the inner iris
float4 innerIrisColor : DIFFUSE
<
    string UIName = "irisInside";
> = {0.1f, 1.0f, 0.1f, 1.0f};
//Color of the outer iris
float4 outerIrisColor : DIFFUSE
<
    string UIName = "irisOutside";
> = {0.1f, 0.1f, 1.0f, 1.0f};
//Color of the veins
float4 veinColor : DIFFUSE
<
    string UIName = "veinColor";
> = {1.0f, 0.1f, 0.1f, 1.0f};

/************** light info **************/ 
float4 ambientColor : Ambient
<
    string UIName = "ambientColor";
> = {0.1f, 0.1f, 0.1f, 1.0f};

float4 light1Pos : POSITION 
< 
	string UIName = "Light 1 Position"; 
	string Object = "PointLight"; 
	string Space = "World"; 
> = {100.0f, 100.0f, 100.0f, 0.0f}; 

float4 light1Color : DIFFUSE 
< 
> = { 1.0f, 1.0f, 1.0f, 1.0f}; 

//---------------------------------- 

// input from application, to vertex shader
struct a2v { 
	float4 position		: POSITION0; 
	float2 texCoord		: TEXCOORD0; 
	float3 normal		: NORMAL0; 
	float3 binormal		: BINORMAL;
	float3 tangent		: TANGENT;
}; 

// output from vertex shader, to fragment program 
struct v2f { 
	float4 position    		: POSITION; 
	float2 texCoord			: TEXCOORD0;
	float3 lightVec    		: TEXCOORD1; 
	float3 eyeVec	    	: TEXCOORD2; 
	float3 worldNormal   	: TEXCOORD3;
	float3 worldBinormal	: TEXCOORD4;
	float3 worldTangent		: TEXCOORD5;

}; 

//----------------------------------------------
//vertex shader
v2f v(a2v In) 
{ 
	v2f Out = (v2f)0; 
    Out.position = mul(In.position, wvp);	//transform vert position to homogeneous clip space 
	
	Out.texCoord.xy = In.texCoord.xy * 2;	//pass through UV coordinates

    float3 worldSpacePos = mul(In.position, world);	//world space position of vertex
    Out.lightVec = light1Pos - worldSpacePos; //world space light vector (from light, to vertex position)
    Out.eyeVec = viewInv[3] - worldSpacePos; //eye vector in world space (from eye, to vertex position)
    
    Out.worldNormal = mul(In.normal, worldIT).xyz;		//compute world space normal 
    Out.worldBinormal = mul(In.binormal, worldIT).xyz;	//compute world space binormal 
    Out.worldTangent = mul(In.tangent, worldIT).xyz;
    return Out; 
}  

//Ambient Pass Pixel Shader
float4 af(v2f In) : COLOR 
{ 
	//normalize our vertex-shader inputs
	float3 L = normalize(In.lightVec.xyz);
	
	float3 Nn = In.worldNormal;
    float3 Tn = In.worldTangent; 
    float3 Bn = In.worldBinormal;
	
	//find N
	float4 normalTexture = tex2D(texSamplerNormal, In.texCoord);
	float3	normal = normalTexture.rgb;
	normal = normal * 2 - 1;
    normal = float4((Nn * normal.z) + (normal.x * Tn + normal.y * -Bn), 1.0);
	float3	N = normalize(normal.rgb);
	
	//red is veins, green is inner iris, blue is outer iris
	float4	rgbTexture = tex2D(texSamplerRGB, In.texCoord);
		
	float4	veins = saturate(rgbTexture.r + veinColor);
    float4	innerIris = saturate(rgbTexture.g + innerIrisColor);
    float4	outerIris = saturate(rgbTexture.b + outerIrisColor);
    float4	eyeColor = veins * innerIris * outerIris;
    
    //overlay map choice
    //RGBA corresponds to 0, 1, 2, and 3 from the "alphaPicker" variable
    float4	overlayTexture = tex2D(texSamplerOverlay, In.texCoord);
    
	float ifOne;
	if (alphaPicker < 1) ifOne = overlayTexture.r; else ifOne = overlayTexture.g; 
	float ifTwo;
	if (alphaPicker < 2) ifTwo = ifOne; else ifTwo = overlayTexture.b; 
	float ifThree;
	if (alphaPicker < 3) ifThree = ifTwo; else ifThree = overlayTexture.a; 
	float3 overlayColor = ifThree;
	
	//the "diffuse" eye color
	float3 eyeComp = eyeColor * overlayColor;
	
	//cube map
	float4	reflTexture = texCUBE(texSamplerRefl, N);
	reflTexture /= reflContrast;

	float3	 ambient = ambientColor * eyeComp + reflTexture;
	float4	ret =  float4(ambient.rgb, 1);
	
	return ret; 
}

//Multi-Pass Pixel Shader
float4 f(v2f In) : COLOR 
{ 
	//normalize our vertex shader variables
	float3 L = normalize(In.lightVec.xyz);		//normalized light vector 
    float3 E = normalize(In.eyeVec.xyz);		//normalized eye vector 
	float3 H = normalize(L + E);				//Normalized half angle vector
	
	float3 Nn = In.worldNormal;
    float3 Tn = In.worldTangent; 
    float3 Bn = In.worldBinormal;
	
	//find N for the iris, this is an inverted version of the normal map
	float4 normalTexture = tex2D(texSamplerNormal, In.texCoord);
	float3	normalIris = float3(((normalTexture.r * -1) + 1), ((normalTexture.g * -1) + 1), 1);
	normalIris = normalIris * 2 - 1;
    normalIris = float4((Nn * normalIris.z) + (normalIris.x * Tn + normalIris.y * -Bn), 1.0);
	float3	Ni = normalize(normalIris.rgb);
	
	//find N for the cornea, this is the normal map (which should bulge out)
	float3	normalCornea = normalTexture.rgb;
	normalCornea = normalCornea * 2 - 1;
    normalCornea = float4((Nn * normalCornea.z) + (normalCornea.x * Tn + normalCornea.y * -Bn), 1.0);
	float3	Nc = normalize(normalCornea.rgb);
	
	//red is veins, green is inner iris, blue is outer iris
	float4	rgbTexture = tex2D(texSamplerRGB, In.texCoord);
		
	float4	veins = saturate(rgbTexture.r + veinColor);
    float4	innerIris = saturate(rgbTexture.g + innerIrisColor);
    float4	outerIris = saturate(rgbTexture.b + outerIrisColor);
    float4	eyeColor = veins * innerIris * outerIris;
    
    //overlay map choice
    //RGBA corresponds to 0, 1, 2, and 3 from the "alphaPicker" variable
    float4	overlayTexture = tex2D(texSamplerOverlay, In.texCoord);
    
	float ifOne;
	if (alphaPicker < 1) ifOne = overlayTexture.r; else ifOne = overlayTexture.g; 
	float ifTwo;
	if (alphaPicker < 2) ifTwo = ifOne; else ifTwo = overlayTexture.b; 
	float ifThree;
	if (alphaPicker < 3) ifThree = ifTwo; else ifThree = overlayTexture.a; 
	float3 overlayColor = ifThree;
	
	//the "diffuse" eye color
	float3 eyeComp = eyeColor * overlayColor;;
	
	//The modulated cube map gives us a cheap image based lighting that breaks up our specular
	float4	reflTexture = texCUBE(texSamplerRefl, Nc);
	reflTexture = pow(reflTexture + reflBrightness, reflContrast);
	
    //diffuse
    //amount of lighting controlled by diffuseBias
	float	NdotL = saturate(dot(Ni, L));
	float3	diffuseColor = eyeComp * NdotL;
	float3	diffuseComp = saturate(diffuseColor + diffuseBias);
	
	//Specular calculations for the cornea
	//This is the tighter specular highlight, broken up by the reflection map
	float	NdotHc = saturate(dot(Nc, H));
	float	specLevelc = pow(NdotHc, glossCornea);
	float4	specCompc = specLevelc * reflTexture;
	
	//Specular calculations for the iris
	//This is a broader highlight on the iris that takes its color
	float	NdotHi = saturate(dot(Ni, H));
	float	specLeveli = pow(NdotHi, glossIris);
	float3	specCompi = specLeveli * normalTexture.a * eyeComp.rgb;
	
	float3	specularComp = specCompc + specCompi;
	
	//final addition
	float3	diffuseSpec = (specularComp + diffuseComp) * light1Color;
	float4	ret =  float4(diffuseSpec, 1);		//Sets w value of the float4 to 1 (a float 4 must be the output of the pixel shader)	
	
	return ret; 
} 

technique Complete 
{  
	pass envPass
    {		 
	VertexShader = compile vs_2_0 v(); 
	PixelShader = compile ps_2_0 af(); 
	} 

	pass light1  
    {		 
	VertexShader = compile vs_2_0 v(); 
	ZEnable = true; 
	ZWriteEnable = false; 
	ZFunc = LessEqual;
	CullMode = none; 
	AlphaBlendEnable = true; 
	SrcBlend = One;
	DestBlend = One;
	PixelShader = compile ps_2_0 f(); 
	}

}  
